/*
  dxf2povprism: Converts a DXF file to linear_spline suitable for use
  in povray. Uses the dime library to read the DXF file. Small fragments
  from dime have been reused here.
  Copyright (C) 2005 Blaise Gassend <blaise at gassend dot com>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <dime/Input.h>
#include <dime/Output.h>
#include <dime/Model.h>
#include <dime/State.h>
#include <dime/Base.h>
#include <dime/entities/Entity.h>
#include <dime/entities/Polyline.h>
#include <dime/entities/LWPolyline.h>
#include <dime/entities/Circle.h>
#include <dime/entities/Vertex.h>
#include <stdio.h>
#include <dime/convert/convert.h>
#include <dime/convert/layerdata.h>
                     
FILE *out = stdout;
FILE *errout = stderr;

double tolerance = 1;
double maxbulge = .1;
bool start_shape = true;
double shape_first_x, shape_first_y;
char *layer = NULL;
bool first_pass = true;
int numpoints = 0;
int numsubdivisions = 0;

void out_pt(dxfdouble x, dxfdouble y)
{
  if (start_shape)
  {
    start_shape = false;
    shape_first_x = x;
    shape_first_y = y;
  }

  if (!first_pass)
    fprintf(out, ",< %g , %g >\n", x, y); 
  numpoints++;
}

void end_shape()
{
  out_pt(shape_first_x, shape_first_y);
  start_shape = true;
  if (!first_pass)
    fprintf(out, "\n");
}

void convert_arc(dxfdouble x1, dxfdouble y1, dxfdouble x2, dxfdouble y2, dxfdouble bulge)
{
  dxfdouble hdx = (x2 - x1) / 2;
  dxfdouble hdy = (y2 - y1) / 2;
  dxfdouble hl = sqrt(hdx * hdx + hdy * hdy);

  if (hl == 0)
  {
    if (errout) 
      fprintf(errout, "Warning: Found a zero length arc, skipping.\n");
    return;
  }
  
  if (fabs(hl * bulge / tolerance) > 1 || fabs(bulge) > maxbulge)
  { // Need to subdivide.
    dxfdouble bulge2 = (sqrt(1 + bulge * bulge) - 1) / bulge; 
    dxfdouble x3 = x1 + hdx + hdy * bulge;
    dxfdouble y3 = y1 + hdy - hdx * bulge;
  
    convert_arc(x1, y1, x3, y3, bulge2);
    convert_arc(x3, y3, x2, y2, bulge2);

    numsubdivisions++;
  }
  else
  { // Draw a line.
    out_pt(x1, y1); 
  }
}

void convert_lwpolyline(dimeLWPolyline *pline)
{
  const dxfdouble *x = pline->getXCoords();
  const dxfdouble *y = pline->getYCoords();
  const dxfdouble *b = pline->getBulges();
      
  int n = pline->getNumVertices();
  bool closed = pline->getFlags() & 1;
  int stop = closed ? n : n-1;

//  if (errout) fprintf(errout, "LWPolyline (%s) with %i points:\n", 
//      closed ? "closed" : "open", n);
 
  if (!closed)
    if (errout) fprintf(errout, "Warning: Unclosed polyline!\n");

  if (stop == 0)
  {
    if (errout) fprintf(errout, "Warning: Only one vertex in polyline!\n");
    return;
  }
  
  for (int i = 0; i < stop; i++) 
  {
    int next = (i+1) % n;
//    if (errout) fprintf(errout, "  <%g, %g> bulge %g\n", x[i], y[i], b[i]);
    convert_arc(x[i], y[i], x[next], y[next], b[i]);
  }

  if (closed) end_shape();
}

void convert_polyline(dimePolyline *pline)
{
  int n = pline->getNumCoordVertices();
  bool closed = pline->getFlags() & 1;
  int stop = closed ? n : n-1;

//  if (errout) fprintf(errout, "Polyline (%s) with %i points:\n", 
//      closed ? "closed" : "open", n);
  
  if (!closed)
    if (errout) fprintf(errout, "Warning: Unclosed polyline!\n");
  
  if (stop == 0)
  {
    if (errout) fprintf(errout, "Warning: Only one vertex in polyline!\n");
    return;
  }
  
  for (int i = 0; i < stop; i++) 
  {
    dimeVertex *v = pline->getCoordVertex(i);
    dimeVec3f pt = v->getCoords();
    dimeParam bulge_param;
    dxfdouble bulge;
    if (v->getRecord(42, bulge_param))
      bulge = bulge_param.double_data;
    else
      bulge = 0;
//    if (errout) fprintf(errout, "  <%g, %g> bulge %g\n", pt[0], pt[1], bulge);
    int next = (i+1) % n;
    dimeVertex *v2 = pline->getCoordVertex(next);
    dimeVec3f pt2 = v2->getCoords();
    convert_arc(pt[0], pt[1], pt2[0], pt2[1], bulge);
  }

  if (closed) end_shape();
}

void convert_circle(dimeCircle *circ)
{
  dimeVec3f center = circ->getCenter();
  dxfdouble radius = circ->getRadius();

//  if (errout) fprintf(errout, "Circle (%g, %g) R=%g\n", center[0], center[1], radius);
  convert_arc(center[0] + radius, center[1], center[0] - radius, center[1], 1);
  convert_arc(center[0] - radius, center[1], center[0] + radius, center[1], 1);
  
  end_shape();
}

bool mycallback(const dimeState *const state, dimeEntity *entity, void *data)
{
  char *Types[] = {
    "Type0",
    "Base",
    "Record",
    "StringRecord",
    "FloatRecord",
    "DoubleRecord",
    "Int8Record",
    "Int16Record",
    "Int32Record",
    "HexRecord",
    "RecordHolder",
    "Class",
    "UnknownClass",
    "Object",
    "UnknownObject",
    "Entity",
    "UnknownEntity",
    "Polyline",
    "Vertex",
    "FaceEntity",
    "ExtrusionEntity",
    "3DFace",
    "Solid",
    "Trace",
    "Line",
    "Point",
    "Block",
    "Insert",
    "Circle",
    "Arc",
    "LWPolyline",
    "Ellipse",
    "Spline",
    "Section",
    "UnknownSection",
    "EntitiesSection",
    "BlocksSection",
    "TablesSection",
    "HeaderSection",
    "ClassesSection",
    "ObjectsSection",
    "Table",
    "TableEntry",
    "UnknownTable",
    "UCSTable",
    "LayerTable",
    "LastTag"
  };

  if (entity->typeId() >= sizeof(Types))
  {
    if (errout) fprintf(errout, "Warning: Entity has invalid type!!\n");
    return true;
  }
    
  if (layer && strcmp(layer, entity->getLayerName()))
    return true; // Not in a layer we are interested in.
  if (!layer)
    if (errout) 
      fprintf(errout, "Found object in layer %s\n", entity->getLayerName());
  
  switch (entity->typeId())
  {
    case dimeBase::dimeLWPolylineType:
      convert_lwpolyline((dimeLWPolyline *) entity);
      break;
    case dimeBase::dimePolylineType:
      convert_polyline((dimePolyline *) entity);
      break;
    case dimeBase::dimeCircleType:
      convert_circle((dimeCircle *) entity);
      break;
    default:
      if (errout) fprintf(errout, "Warning: ignoring entity of type %s\n", Types[entity->typeId()]);
      break;
  }
  return true;
}

void usage(char *progname)
{           
  if (errout) fprintf(errout,
      "Usage: %s [-i infile] [-o outfile] [-t tol] [-b bulge_err] [-l layer]\n"
      "(default infile is stdin, default outfile is stdout)\n\n"
      "Options:\n"
      "-t <tol>     Maximum error when tessellating curves\n"
      "-b <numsub>  Maximum angular error when tessellating curves\n"
      "-l <layer>   Name of the layer to use, default is all layers\n"
      "-q           Suppress most messages to stderr\n"
      "Please send bugs/comments to gassend@alum.mit.edu\n\n",
      progname);

  exit(1);
}

int main(int argc, char **argv)
{
  char *infile = NULL;
  char *outfile = NULL;
  dimeInput in;
  
  for (int i = 1; i < argc; i++)
  {
    if (argv[i][0] == '-')
    {
      switch (argv[i][1])
      {
	case 'q':
	  errout = NULL;
	  continue;
      }
      
      i++;
      if (i >= argc)
      {
	if (errout) fprintf(errout, "-%c expecting an option.\n", argv[i-1][1]);
	usage(argv[0]); // No argument follows the current argument.
      }
      
      switch (argv[i - 1][1])
      {
	case 'i':
	  infile = argv[i];
	  break;
	case 't':
	  tolerance = atof(argv[i]);
	  break;
	case 'b':
	  maxbulge = atof(argv[i]);
	  break;
	case 'l':
	  layer = argv[i];
	  break;
        case 'o':
	  outfile = argv[i];
	  break;
	default:
	  if (errout) fprintf(errout, "Unknown option %c.\n", argv[i-1][1]);
	  usage(argv[0]);
      }
    }
  }
  
  if (infile)
  {
    if (!in.setFile(infile))
    {
      if (errout) fprintf(errout, "Error opening file %s for reading.\n", infile);
      return 1;
    }
  }
  else 
  {
    if (errout) fprintf(errout, "Reading from stdin.\n");
    if (!in.setFileHandle(stdin))
    {
      if (errout) fprintf(errout, "Error opening stdin for reading.\n");
      return 1;
    }
  }
 
  if (outfile && !(out = fopen(outfile, "wb")))
  {
    if (errout) fprintf(errout, "Error opening file %s for writing.\n", outfile);
    return 1;
  }

  dimeModel model;

  if (!model.read(&in))
  {
    if (errout) fprintf(errout, "Error reading file.\n");
    return 1;
  }

  first_pass = true;
  model.traverseEntities(mycallback, NULL, false, true, false);
  if (errout) fprintf(errout, "Number of subdivisions: %i\n", numsubdivisions);
  if (errout) fprintf(errout, "Number of vertices: %i\n", numpoints);
  fprintf(out, "%i\n", numpoints);
  first_pass = false;
  model.traverseEntities(mycallback, NULL, false, true, false);
  
  if (out != stdout) fclose(out);
  return 0;
}
